
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <algorithm>
#include <vector>
#include "focusStacking.h"
#include "settings.h"
#include "../Include/dof.h"

using namespace std;

/*
 *  Constructor for focus stacking.
 */
FocusStacking::FocusStacking(
    int nImages,
    double ***measures,
    Settings *settings )
{
    this->nImages = nImages;
    this->measures = measures;
    this->settings = settings;
    nRows = settings->getRows();
    nCols = settings->getCols();
    debug = settings->getDebug();
    t1 = settings->getThreshold1();
    /*
     * For narrow apertures, t2 is multiplied by a weight
     * that was found by searching for the multiplier that
     * gave the best results within the evaluation framework.
     */
    if( settings->getAperture() < 8.0 ) {
	t2 = settings->getThreshold2();
    }
    else {
	t2 = settings->getThreshold2() * settings->getWeight();
    }
    t3 = settings->getThreshold3();

    dof = new DepthOfField(
	settings->getCoC(),
	settings->getLens(),
	settings->getAperture()
    );

    cover = new bool*[8*nRows*nCols];
    for( int i = 0; i < 8*nRows*nCols; i++ ) {
	cover[i] = new bool[256];
	for( int k = 0; k < 256; k++ ) {
	    cover[i][k] = false;
	}
    }

    depthMap = new int*[nRows];
    for( int i = 0; i < nRows; i++ ) {
	depthMap[i] = new int[nCols];
	for( int j = 0; j < nCols; j++ ) {
	    depthMap[i][j] = -1;
	}
    }
}

/*
 *  Destructor for focus stacking.
 */
FocusStacking::~FocusStacking()
{
    for( int i = 0; i < nRows; i++ ) {
	delete[] depthMap[i];
    }
    delete[] depthMap;

    for( int i = 0; i < 8*nRows*nCols; i++ ) {
	delete[] cover[i];
    }
    delete[] cover;
}

double
FocusStacking::getMeasure( int i, int j, int k )
{
    if( (0 <= i) && (i < nRows) &&
        (0 <= j) && (j < nCols) ) {
	return( measures[i][j][k] );
    }
    else {
	return( 0.0 );
    }
}

double
FocusStacking::getCellMeasure( int i, int j, int k )
{
    double fm =
	getMeasure( i-1, j-1, k ) +
	getMeasure( i-1, j  , k ) +
	getMeasure( i-1, j+1, k ) +
	getMeasure( i  , j-1, k ) +
	getMeasure( i  , j  , k ) +
	getMeasure( i  , j+1, k ) +
	getMeasure( i+1, j-1, k ) +
	getMeasure( i+1, j  , k ) +
	getMeasure( i+1, j+1, k );

    return( fm );
}

/*
 *  Construct the set covering instance for choosing
 *  the images for focus stacking.
 */
void
FocusStacking::constructSetCoveringInstance()
{
    int withinRow[nRows][256];
    for( int i = 0; i < nRows; i++ ) {
	for( int k = 0; k < 256; k++ ) {
	    withinRow[i][k] = 0;
	}
    }

    /*
     *  Fill in depth map with max of focus measure for that cell.
     */
    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {

	double fm[256];
        if( settings->getSelectionMethod() == 0 ) {
	    /*
	     *  Vaquero et al.
	     */
	    for( int k = 0; k < nImages; k++ ) {
	        fm[k] = measures[i][j][k];
	    }
	}
	else {
	    /*
	     *  Adaptive: Smoothing, overlapping grid.
	     */
	    for( int k = 0; k < nImages; k++ ) {
	        fm[k] = getCellMeasure( i, j, k );
	    }
	}

	double max = 0.0;
	for( int k = 0; k < nImages; k++ ) {
	    if( max < fm[k] ) {
	        max = fm[k];
	    }
	}
	for( int k = 0; k < nImages; k++ ) {
	    if( max == fm[k] ) {
		depthMap[i][j] = k;
		if( k != 0 ) withinRow[i][k-1]++;
		withinRow[i][k]++;
		withinRow[i][k+1]++;
		break;	// take first max closest to camera
	    }
	}
    }

    if( debug ) {
	printf( "\nDepth map before reliability check\n\n" );
	for( int i = 0; i < nRows; i++ ) {
	    for( int j = 0; j < nCols; j++ ) {
		if( depthMap[i][j] == -1 ) printf( "   -" );
		else			   printf( "%4d", depthMap[i][j] );
	    }
	    printf( "\n" );
	}
	printf( "\n" );
    } // debug

    /*
     *  Predict which entries in the depth map are unreliable
     *  because of a lack of contrast in the sequence of images
     *  for that cell.
     */
    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {

	double fm[256];
        if( settings->getSelectionMethod() == 0 ) {
	    /*
	     *  Vaquero et al.
	     */
	    for( int k = 0; k < nImages; k++ ) {
	        fm[k] = measures[i][j][k];
	    }
	}
	else {
	    /*
	     *  Adaptive: Smoothing, overlapping grid.
	     */
	    for( int k = 0; k < nImages; k++ ) {
	        fm[k] = getCellMeasure( i, j, k );
	    }
	}

	int sameInRowCount = 0;
	for( int k = 0; k < nCols; k++ ) {
	    if( depthMap[i][j] == depthMap[i][k] ) {
		sameInRowCount++;
	    }
	}

	/*
	 *  Determine the mean.
	 */
	double mean = 0.0;
	for( int k = 0; k < nImages; k++ ) {
	    mean += fm[k];
	}
	mean = mean / nImages;

	/*
	 *  Determine the max.
	 */
	double max = 0.0;
	for( int k = 0; k < nImages; k++ ) {
	    if( max < fm[k] ) {
	        max = fm[k];
	    }
	}

	/*
	 *  Determine the standard deviation and kurtosis.
	 */
	double standard_deviation = 0.0;
	double kurtosis = 0.0;
	for( int k = 0; k < nImages; k++ ) {
	    standard_deviation += pow(fm[k] - mean, 2.0);
	    kurtosis += pow(fm[k] - mean, 4.0);
	}
	kurtosis = kurtosis / nImages;
	kurtosis = kurtosis / pow(standard_deviation/nImages, 2.0);
	standard_deviation = sqrt(standard_deviation / nImages);

        bool reliable;
        if( settings->getSelectionMethod() == 0 ) {
            /*
             *  Vaquero heuristic and Vaquero method.
             */
            reliable = (standard_deviation > t1);
	}
	else {
            /*
             *  Machine learning heuristic and DOF method.
             */
            reliable = (kurtosis > t3);
        }

	if( !reliable ) {
	    if( debug ) {
		printf( "cell (%d, %d), no peak found at %d, kurtosis = %8.5f\n",
		    i, j, depthMap[i][j], kurtosis );
	    }
	    depthMap[i][j] = -1;
	}
	else {
            if( debug ) {
                printf( "cell (%d, %d), peak found at %d, kurtosis = %8.5f\n",
                    i, j, depthMap[i][j], kurtosis );
            }
	}
    }

    if( debug ) {
	printf( "\nDepth map after reliability check\n\n" );
	for( int i = 0; i < nRows; i++ ) {
	    for( int j = 0; j < nCols; j++ ) {
		if( depthMap[i][j] == -1 ) printf( "   -" );
		else			   printf( "%4d", depthMap[i][j] );
	    }
	    printf( "\n" );
	}
	printf( "\n" );
    } // debug

    /*
     *  Fill in the (0,1)-matrix representing the set covering instance.
     *  The rows represent objects/cells, the columns represent images.
     *  There is a 1 in row i, column j if object i is in acceptably
     *  sharp focus in image j.
     */
    nObjects = 0;
    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {

	int peak = depthMap[i][j];
	if( peak == -1 ) {
	    continue;
	}

	double fm[256];
        if( settings->getSelectionMethod() == 0 ) {
	    /*
	     *  Vaquero et al.
	     */
	    for( int k = 0; k < nImages; k++ ) {
	        fm[k] = measures[i][j][k];
	    }
	}
	else {
	    /*
	     *  Adaptive.
	     */
	    for( int k = 0; k < nImages; k++ ) {
	        fm[k] = getCellMeasure( i, j, k );
	    }
	}

	double max = fm[peak];
	double mean = 0.0;
	for( int k = 0; k < nImages; k++ ) {
	    mean += fm[k];
	}
	mean = mean / nImages;

	if( settings->getSelectionMethod() == 0 ) {
	    /*
	     *  Vaquero heuristic and Vaquero method.
	     */
            if( debug ) {
                printf( "cell (%d, %d), peak found at %d\n", i, j, peak );
            }
            cover[nObjects][peak] = true;
            for( int p = peak+1; (p < nImages) && ((max - fm[p]) < t2); p++ ) {
                cover[nObjects][p] = true;
            }
            for( int p = peak-1; (p >= 0) && ((max - fm[p]) < t2); p-- ) {
                cover[nObjects][p] = true;
            }
            nObjects++;
	}
        else {
            /*
             *  Machine learning heuristic and DOF method.
             */
            int nearPosition;
            int farPosition;
            dof->dofInLensPositions(
                peak, // lensPosition
                nearPosition,
                farPosition
            );
            if( farPosition >= nImages ) {
               farPosition = nImages-1;
            }
            for( int p = nearPosition; p <= farPosition; p++ ) {
                cover[nObjects][p] = true;
            }
            if( debug ) {
                printf( "range: [%d,... %d,... %d]\n",
                    nearPosition, peak, farPosition );
            } // debug
            nObjects++;
        } // else
    } // for i, j rows and columns

    /*
     *  Find "off-peaks", improve robustness of set
     *  covering solution.
     */
    if( settings->getSelectionMethod() != 0 ) {

        int peaks[256];
        for( int k = 0; k <= dof->maxLensPosition(); k++ ) {
            peaks[k] = 0;
        }
        for( int i = 0; i < nRows; i++ )
        for( int j = 0; j < nCols; j++ ) {
            if( depthMap[i][j] >= 0 ) {
                peaks[depthMap[i][j]]++;
            }
        }

        int width[256];
        for( int k = 0; k <= dof->maxLensPosition(); k++ ) {
            width[k] = 0;
        }
        for( int k = 0; k <= dof->maxLensPosition(); k++ ) {
	    int l, r;
	    for( l = k; l >= 0 && peaks[l]; l-- ) ;
	    for( r = k; r <= dof->maxLensPosition() && peaks[r]; r++ ) ;
	    width[k] = r - l - 1;
	}

        bool offpeaks[256];
        for( int k = 0; k <= dof->maxLensPosition(); k++ ) {
            offpeaks[k] = false;
        }
        for( int k = 2; k < dof->maxLensPosition(); k++ ) {
	    if( (peaks[k] == 0 && peaks[k-1] > 0) ||
	        (peaks[k] == 0 && peaks[k+1] > 0) ) {
		offpeaks[k] = true;
	    }
	}

        for( int k = 2; k < dof->maxLensPosition(); k++ ) {

	    if( !offpeaks[k] ) continue;

            int nearPosition;
            int farPosition;
            dof->dofInLensPositions(
                k, // lensPosition
                nearPosition,
                farPosition
            );
            if( farPosition >= nImages ) {
               farPosition = nImages-1;
            }
            for( int p = nearPosition; p <= farPosition; p++ ) {
                cover[nObjects][p] = true;
            }
            if( debug ) {
                printf( "offpeak range: [%d,... %d,... %d]\n",
                    nearPosition, k, farPosition );
            } // debug
            nObjects++;
	}

        if( debug ) {
	    printf( "\n" );
	    for( int k = 0; k <= dof->maxLensPosition(); k++ ) {
	        if( peaks[k] != 0 ) {
		    printf( "peaks[%2d] = %d (width = %d)\n",
			k, peaks[k], width[k] );
	        }
	    }
	    for( int k = 0; k <= dof->maxLensPosition(); k++ ) {
	        if( offpeaks[k] ) {
		    printf( "offpeaks[%2d] = %d\n", k, offpeaks[k] );
	        }
	    }
        }
    }

    /*
     *  Print set covering (0,1)-matrix.
     */
    if( debug ) {
	printf( "\nSet covering (0,1)-matrix:\n\n" );
	for( int i = 0; i < nObjects; i++ ) {
	    for( int k = 0; k < nImages; k++ ) {
	        printf( "%d", (int)cover[i][k] );
	    }
	    printf( "\n" );
	}
	printf( "\n" );
	printf( "nRows * nCols = %d, nObjects = %d\n", nRows*nCols, nObjects ); 
    } // debug
}

/*
 *  Select images to stack by solving the set covering instance.
 */
vector<int>
FocusStacking::chooseAdaptiveAutofocus( const vector<int> &images )
{
    vector<int> selectedImages;
    bool processed[8*nRows*nCols];

    for( int i = 0; i < nObjects; i++ ) {
	processed[i] = false;
    }

    /*
     *  Reduction rule for rows: remove any row that is a
     *  superset of another row.
     */
    if( settings->getSelectionMethod() != 0 ) {
        int sr[8*nRows*nCols];
        int er[8*nRows*nCols];
        for( int i = 0; i < nObjects; i++ ) {
            unsigned int k;
            for( k = 0; (cover[i][images[k]] == 0) && (k < images.size()); k++ ) {
                ;
            }
            sr[i] = k;
            for( k = images.size()-1; (cover[i][images[k]] == 0) && (k > 0); k-- ) {
                ;
            }
            er[i] = k;
        }
        for( int i = 0; i < nObjects; i++ )
        for( int j = 0; j < nObjects; j++ ) if( i != j ) {
            if( !processed[j] ) {
                /*
                 *  If row i is a superset, remove it.
                 */
                if( (sr[i] <= sr[j]) && (er[i] >= er[j]) ) {
                    processed[i] = true;
		    if( debug ) {
                        printf("Row %d is removed due to row %d\n", i, j );
		    } // debug
                    break;
                }
            }
        }
    } // if

    /*
     *  Sweep through images.
     */
    for( unsigned int k = 1; k < images.size(); k++ ) {
        int pl = images[k-1];
        int l = images[k];
        /*
         *  For each object not yet processed ...
         */
        for( int i = 0; i < nObjects; i++ ) if( !processed[i] ) {
            if( !cover[i][l] && cover[i][pl] ) {
		/*
		 *  Add image to set of selected images.
		 */
                selectedImages.push_back( pl );
		if( debug ) {
		    printf( "Selecting image %d\n", pl );
		}
		/*
		 *  Mark any newly covered objects as processed.
		 */
                for( int i2 = 0; i2 < nObjects; i2++ ) {
                    if( cover[i2][pl] ) {
                        processed[i2] = true;
                    }
                }
            }
        }
    }

    /*
     *  Special case at last image.
     */
    int l = images[images.size() - 1];
    for( int i = 0; i < nObjects; i++ ) if( !processed[i] ) {
        if( cover[i][l] ) {
            selectedImages.push_back( l );
	    if( debug ) {
	        printf( "Selecting image %d\n", l );
	    }
            for( int i2 = 0; i2 < nObjects; i2++ ) {
                if( cover[i2][l] ) {
                    processed[i2] = true;
                }
            }
        }
    }

    return( selectedImages );
}

/*
 *  Remove images that do not need to be looked at.
 */
vector<int>
FocusStacking::shrinkImageSet()
{
    vector<int> imagesToExamine;

    /*
     *  Reduction rule for columns: remove any column that is a
     *  subset of another column.
     */
    bool include[nImages];
    for( int i = 0; i < nImages; i++ ) {
	include[i] = true;
    }

    if( settings->getSelectionMethod() != 0 ) {
        for( int i = 0; i < nImages; i++ )
        for( int j = 0; j < nImages; j++ ) if( i != j ) {
            if( include[j] ) {
                /*
                 *  If column i is a subset of column j, remove column i.
                 */
                bool subset = true;
                for( int k = 0; k < nObjects && subset; k++ ) {
                    if( cover[k][i] > cover[k][j] ) {
                        subset = false;
                    }
                }
                if( subset ) {
                    include[i] = false;
		    if( debug ) {
                        printf("Column %d is removed due to column %d\n", i, j );
		    } // debug
                    break;
                }
            }
        }
    } // if

    for( int k = 0; k < nImages; k++ ) {
	if( include[k] ) {
	    if( debug ) {
	        printf( "image to consider: %d\n", k );
	    } // debug
	    imagesToExamine.push_back( k );
	}
    }

    if( debug ) {
        printf( "total images to consider: %d\n", imagesToExamine.size() );
    } // debug

    return( imagesToExamine );
}

/*
 *  Select images to stack using new adaptive autofocus algorithm.
 */
vector<int>
FocusStacking::selectAdaptiveAutofocus()
{
    /*
     *  Select high resolution images to acquire by constructing
     *  and solving set covering instance.
     */
    constructSetCoveringInstance();
    vector<int> selectedImages;
    selectedImages = chooseAdaptiveAutofocus( shrinkImageSet() );

    /*
     *  Determine if acquiring an image at the hyperfocal distance
     *  can be used to replace two or more images scheduled to be
     *  acquired.
     */
    if( settings->getSelectionMethod() != 0 ) {
        int lensPosition;
        int nearPosition;
        int farPosition;
        dof->hyperfocalInLensPositions(
            lensPosition,
            nearPosition,
            farPosition
        );

	int selectedNearPosition[256];
	int selectedFarPosition[256];
	int subsumed = 0;
	for( unsigned int i = 0; i < selectedImages.size(); i++ ) {
	    dof->dofInLensPositions(
	        selectedImages[i],
	        selectedNearPosition[i],
	        selectedFarPosition[i]
	    );
	    if( (nearPosition <= selectedNearPosition[i]) &&
	        (selectedFarPosition[i] <= farPosition) ) {
		subsumed++;
	    }
	}
	if( subsumed > 1 ) {
            int k = selectedImages.size() - 1;
	    if( (k > 0 ) &&
		(nearPosition <= selectedNearPosition[k]) &&
	        (selectedFarPosition[k] <= farPosition) ) {
		k--;
	    }
	    if( debug ) {
		printf( "Subsumed = %d\n", subsumed );
		printf( "Hyperfocal: %d <= %d <= %d\n",
		    nearPosition, lensPosition, farPosition );
		printf( "Replacing " );
		for( unsigned int l = k; l < selectedImages.size(); l++ ) {
		    printf( "%d ", selectedImages[l] );
		}
		printf( "with hyperfocal distance image %d\n", lensPosition );
	    }
            selectedImages[k] = lensPosition;
            selectedImages.resize( k+1 );
	}
    }

    printf( "\nCovered images:\n" );
    for( unsigned int i = 0; i < selectedImages.size(); i++ ) {
	int nearPosition;
	int farPosition;

	/*
	 *  Check if at hyperfocal lens position.
	 */
	int hyperfocal;
	dof->hyperfocalInLensPositions(
	    hyperfocal,
	    nearPosition,
	    farPosition
	);
	if( selectedImages[i] != hyperfocal ) {
	    dof->dofInLensPositions(
	        selectedImages[i], // lensPosition
	        nearPosition,
	        farPosition
	    );
	}

	printf ( "%d <= %d <= %d\n",
	    nearPosition, selectedImages[i], farPosition );
    }

    return( selectedImages );
}


