/*
 *  This class takes in a grid of focus measures and can be used
 *  to choose images based on various selection algorithms.
 */
#ifndef _FOCUSSTACKING_H
#define _FOCUSSTACKING_H

#include <vector>

class Settings;
class DepthOfField;

class FocusStacking
{
    public:

	FocusStacking( int nImages, double ***measures, Settings *settings );
	~FocusStacking();

	/*
	 *  Select images to stack using new Adaptive Autofocus
	 *  algorithm.
	 */
	std::vector<int> selectAdaptiveAutofocus();

    private:
	/*
         * Threshold for adaptive method:
	 *    t3 is threshold for predicting whether a depth is reliable.
	 *
	 *  Thresholds for Vaquero method:
	 *    t1 is threshold for foreground/background classification.
	 *    t2 is threshold for sharp/blurry classification.
	 *
	 *    t1 = 0.05	    // original value used in Vaquero paper
	 *    t2 = 0.20	    // original value used in Vaquero paper
	 */
	double t1;
	double t2;
	double t3;

	int nObjects;
	int nImages;

	int nRows;
	int nCols;

	int debug;

	/*
	 *  cover[i][k] == 1 if the i^th object is sharp in image k,
	 *		   0 otherwise.
	 */
	bool **cover;

	/*
	 *  measures[i][j][k] is the focus measure at cell (i,j) in image k.
	 */
	double ***measures;
	double getMeasure( int i, int j, int k );
	double getCellMeasure( int i, int j, int k );

	/*
	 *  depthMap[i][j] is the estimated depth (in lens positions)
	 *  of cell (i,j) or -1 if determined to be unreliable.
	 */
	int **depthMap;

	/*
	 *  Depth of field calculations and mapping to lens positions.
	 */
	DepthOfField *dof;

	/*
	 *  Experimental settings.
	 */
	Settings *settings;

	/*
	 *  Classify every object in every cell in every image as sharp/blurry.
	 *  Must be called before selection.
	 */
	void constructSetCoveringInstance();

	/*
	 *  Remove images that do not need to be looked at.
	 */
	std::vector<int> shrinkImageSet();

	/*
	 *  Given a set of images, performs the adaptive autofocus 
	 *  algorithm (used by selectAdaptiveAutofocus).
	 */
	std::vector<int> chooseAdaptiveAutofocus( const std::vector<int>& );
};

#endif // _FOCUSSTACKING_H
