
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <vector>
#include "../Include/focusMeasure.h"
#include "focusStacking.h"
#include "settings.h"

using namespace std;

void readGray( char *fileName, int n, uchar *buffer );


int
main(int argc, char *argv[] )
{
    if( argc <= 3 ) {
        fprintf( stderr, "Usage: adaptive benchmark settings fileName0.gray ...\n" );
        fprintf( stderr, "\tsettings -- file containing settings\n" );
        fprintf( stderr, "\tfileName0.gray -- file containing gray levels\n" );
        exit( 6 );
    }

    printf( "benchmark = %s\n", argv[1] );
    Settings *settings = new Settings( argv[2] );
    settings->printSettings();

    const int w = 1056;
    const int h =  704;
    const int n = w * h;
    uchar buffer[n];

    int nRows = settings->getRows();
    int nCols = settings->getCols();

    double ***measures = new double**[nRows];
    for( int i = 0; i < nRows; i++ ) {
	measures[i] = new double*[nCols];
	for( int j = 0; j < nCols; j++ ) {
	    measures[i][j] = new double[argc];
	}
    }

    FocusMeasure focus;

    /*
     *  Read in images (gray values) and compute focus measures.
     */
    double v = 0;
    double min = HUGE_VAL;
    double max = 0;
    int gridWidth = w / nCols;
    int gridHeight = h / nRows;
    int apply = settings->getFocusMeasure();
    for( int k = 3; k < argc; k++ ) {
        readGray( argv[k], n, buffer );
        for( int i = 0; i < nRows; i++ )
        for( int j = 0; j < nCols; j++ ) {
	    switch( apply ) {
		case  0: v = focus.squaredGradient(
				buffer,
				w,
				h,
				j * gridWidth,
				i * gridHeight,
				gridWidth,
				gridHeight
			     );
			 break;
		case  1: v = focus.brenner(
				buffer,
				w,
				h,
				j * gridWidth,
				i * gridHeight,
				gridWidth,
				gridHeight
			 );
			 break;
		case  2: v = focus.nokiaMeasure(
				buffer,
				w,
				h,
				j * gridWidth,
				i * gridHeight,
				gridWidth,
				gridHeight
			  );
			  break;
	    }
	    if( max < v ) {
		max = v;
	    }
	    if( min > v ) {
		min = v;
	    }
	    measures[i][j][k-3] = (double)v;
        }
    }

    /*
     *  Smooth the data by performing a weighted moving average.
     */
/*
    double s[256];
    int nn = argc-3;
    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {
	s[0] = (3.0*measures[i][j][0] +
	measures[i][j][1]) / 4.0;
	for( int k = 1; k < nn-1; k++ ) {
	    s[k] = (	measures[i][j][k-1] +
		    2.0*measures[i][j][k  ] +
			measures[i][j][k+1]) / 4.0;
	}
	s[nn-1] = (    measures[i][j][nn-2] +
	           3.0*measures[i][j][nn-1]) / 4.0;
	for( int k = 0; k < nn; k++ ) {
	    measures[i][j][k] = s[k];
	}
    }
*/

    /*
     *  Normalize focus measures.
     */
    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {
	for( int k = 0; k < argc-3; k++ ) {
	    measures[i][j][k] = (measures[i][j][k] - min)/(double)(max - min);
        }
    }

    /*
     *  Choose high resolution images to acquire.
     */
    FocusStacking f( argc-3, measures, settings );
    vector<int> adaptiveAutofocus = f.selectAdaptiveAutofocus();
    printf( "\nSelected images (%d):\n", adaptiveAutofocus.size() );
    for( unsigned int i = 0; i < adaptiveAutofocus.size(); i++ ) {
        printf( "%d\n", adaptiveAutofocus[i] );
    }
    printf( "\n\n" );

    /*
     *  Cleanup.
     */
    delete settings;
    for( int i = 0; i < nRows; i++ ) {
        for( int j = 0; j < nCols; j++ ) {
            delete[] measures[i][j];
        }
        delete[] measures[i];
    }
    delete[] measures;

    return( 0 );
}

/*
 *  Read the gray values from a file into buffer.
 */
void readGray( char *fileName, int n, uchar *buffer )
{
    FILE *fp;

    fp = fopen( fileName, "rb" );
    if( fp == NULL ) {
    	fprintf( stderr, "No such file: %s\n", fileName );
    	exit( 1 );
    }

    int result = fread( buffer, 1, n, fp );
    if( result != n ) {
    	fprintf( stderr, "Wrong number of bytes: %s\n", fileName );
    	exit( 1 );
    }

    fclose( fp );
}

