/*
 *  Routines for storing and retrieving experimental parameters.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "settings.h"


/*
 *  Constructor.
 */
Settings::Settings( char *fileName )
{
    focusMeasure = 0;		// squared gradient
    circleOfConfusion = 0.019;	// APS_C
    lens = 50;			// 50 mm
    aperture = 1.4;		// 1.4, 2.0, 2.8, 4.0, 8.0
    selection = 1;		// 0 = Vaquero, 1 = adaptive 
    filter = 0;			// 0 = none
    nRows = 4;
    nCols = 6;
    debug = 0;
    t1 = 0.0;
    t2 = 0.0;
    t3 = 0.0;
    weight = 1.0;

    initialize( fileName );
}

/*
 *  Destructor.
 */
Settings::~Settings()
{
}

/*
 *  Read in and store the experimental parameters.
 */
void
Settings::initialize( char *fileName )
{
    FILE *fp;
    char line[256];

    fp = fopen( fileName, "r" );
    if( fp == NULL ) {
    	fprintf( stderr, "No such file: %s\n", fileName );
    	exit( 1 );
    }

    while( fgets( line, 256, fp ) != NULL ) {
	if( line[0] == '%' ) continue;
	if( line[0] == '#' ) continue;

	if( strncmp( "focusMeasure", line, 8 ) == 0 ) {
	    sscanf( line, "focusMeasure = %d", &focusMeasure );
	}
	else
	if( strncmp( "circleOfConfusion", line, 8 ) == 0 ) {
	    double v;
	    sscanf( line, "circleOfConfusion = %lf", &v );
	    circleOfConfusion = v;
	}
	else
	if( strncmp( "lens", line, 4 ) == 0 ) {
	    sscanf( line, "lens = %d", &lens );
	}
	else
	if( strncmp( "aperture", line, 8 ) == 0 ) {
	    double v;
	    sscanf( line, "aperture = %lf", &v );
	    if(      v == 1.4 ) aperture = pow( 2.0, 0.5 );
	    else if( v == 2.0 ) aperture = pow( 2.0, 1.0 );
	    else if( v == 2.8 ) aperture = pow( 2.0, 1.5 );
	    else if( v == 4.0 ) aperture = pow( 2.0, 2.0 );
	    else if( v == 8.0 ) aperture = pow( 2.0, 3.0 );
	    else {
		printf( "aperture not found in list: %12.8f\n", v );
		exit( 1 );
	    }
	}
	else
	if( strncmp( "selection", line, 8 ) == 0 ) {
	    sscanf( line, "selection = %d", &selection );
	}
	else
	if( strncmp( "nRows", line, 4 ) == 0 ) {
	    sscanf( line, "nRows = %d", &nRows );
	}
	else
	if( strncmp( "nCols", line, 4 ) == 0 ) {
	    sscanf( line, "nCols = %d", &nCols );
	}
	else
	if( strncmp( "debug", line, 4 ) == 0 ) {
	    sscanf( line, "debug = %d", &debug );
	}
	else
	if( strncmp( "t1", line, 2 ) == 0 ) {
	    sscanf( line, "t1 = %lf", &t1 );
	}
	else
	if( strncmp( "t2", line, 2 ) == 0 ) {
	    sscanf( line, "t2 = %lf", &t2 );
	}
	else
	if( strncmp( "t3", line, 2 ) == 0 ) {
	    sscanf( line, "t3 = %lf", &t3 );
	}
	else
	if( strncmp( "weight", line, 6 ) == 0 ) {
	    sscanf( line, "weight = %lf", &weight );
	}
    }

    fclose( fp );
}

void
Settings::printSettings()
{
    if( focusMeasure == 0 ) {
	printf( "focusMeasure = squared gradient\n" );
    }
    else
    if( focusMeasure == 2 ) {
	printf( "focusMeasure = nokia\n" );
    }
    printf( "circleOfConfusion = %5.3f\n", circleOfConfusion );
    printf( "lens = %d mm\n", lens );
    printf( "aperture = f/%3.1f\n", aperture );
    if( selection == 0 ) {
	printf( "weight scheme = 0\n" );
	printf( "selection scheme = Vaquero\n" );
    }
    else {
	printf( "weight scheme = 1\n" );
	printf( "selection scheme = adaptive\n" );
    }
    printf( "nRows = %d\n", nRows );
    printf( "nCols = %d\n", nCols );
    printf( "t1 = %12.8f\n", t1 );
    printf( "t2 = %12.8f\n", t2 );
    printf( "t3 = %12.8f\n", t3 );
}

int
Settings::getFocusMeasure()
{
    return( focusMeasure );
}

double
Settings::getCoC()
{
    return( circleOfConfusion );
}

int
Settings::getLens()
{
    return( lens );
}

double
Settings::getAperture()
{
    return( aperture );
}

int
Settings::getSelectionMethod()
{
    return( selection );
}

int
Settings::getRows()
{
    return( nRows );
}

int
Settings::getCols()
{
    return( nCols );
}

int
Settings::getDebug()
{
    return( debug );
}

double
Settings::getThreshold1()
{
    return( t1 );
}

double
Settings::getThreshold2()
{
    return( t2 );
}

double
Settings::getThreshold3()
{
    return( t3 );
}

double
Settings::getWeight()
{
    return( weight );
}

