/*
 *  Routines for depth of field calculations and mapping to lens positions.
 */

#include <stdio.h>
#include <math.h>
#include "dof.h"

const double DepthOfField::infinity = 100000.0;	// 100 km

/*
 *  Canon EF 50mm f/1.4 USM.
 */
const int DepthOfField::MAX_LensPositions_50mm = 165;	// 0, ..., 165
const DepthOfField::DataType DepthOfField::data_50mm[] = {    
    { 0,        0.45 },
    { 48,       0.60 },
    { 76,       0.75 },
    { 81,       0.80 },
    { 92,       0.90 },
    { 100,      1.00 },
    { 120,      1.50 },
    { 143,      3.0 },
    { 163,      DepthOfField::infinity },
    { 166,      DepthOfField::infinity },
    { -1, -1 }
};

/*
 *  Canon EF 70-200mm f/2.8 II USM.
 *  Focal length set at 200mm.
 */
const int DepthOfField::MAX_LensPositions_200mm = 230;	// 0, ..., 230
const DepthOfField::DataType DepthOfField::data_200mm[] = {    
    { 0,        1.20 },
    { 10,       1.22 },
    { 57,       1.49 },
    { 61,       1.52 },
    { 104,      2.00 },
    { 112,      2.13 },
    { 146,      3.00 },
    { 150,      3.05 },
    { 176,      4.57 },
    { 180,      5.00 },
    { 198,      9.14 },
    { 200,      10.0 },
    { 224,      DepthOfField::infinity },
    { 231,      DepthOfField::infinity },
    { -1, -1 }
};

DepthOfField::DepthOfField(
    double circleOfConfusion,
    int focalLength,
    double aperture )
{
    this->circleOfConfusion = circleOfConfusion;
    this->focalLength = focalLength;
    this->aperture = aperture;

    if( focalLength == 50 ) {
	nPos = MAX_LensPositions_50mm;
	interpolate( data_50mm, lens );
    }
    else { /* focalLength == 200 */
	nPos = MAX_LensPositions_200mm;
	interpolate( data_200mm, lens );
    }
}

DepthOfField::~DepthOfField()
{
}

int
DepthOfField::maxLensPosition()
{
    return( nPos );
}

/*
 *  From known (lens positions, distance marks) pairings,
 *  interpolate to fill in the remaining values.
 *
 *  See: http://imaginatorium.org/stuff/focus.htm
 */
void
DepthOfField::interpolate( const DataType data[], double lens[] )
{
    int i, j;
    int a, b;
    double step, dist;

    for( i = 0; data[i].position != -1; i++ ) {    
	a = data[i].position;
	b = data[i+1].position;

	dist = 1.0/data[i].distance;
	step = (1.0/data[i+1].distance - 1.0/data[i].distance) /
			(double)(b - a);
	//printf("a = %d, b = %d, dist = %12.8f, step = %12.8f\n",
	    //a,b,dist,step);
	for( j = a; j < b; j++ ) {    
	    //printf("j = %d, a = %d, b = %d, dist = %12.8f, step = %12.8f\n",
		//j,a,b,dist,step);
	    lens[j] = 1.0/dist;
	    dist += step;
	    //printf( "lens[%d] = %12.8f\n", j, lens[j] );
	}
    }
    lens[nPos] = DepthOfField::infinity;
}

/*
 *  Determine the depth of field in meters.
 *  Depth of field calculations are in millimetres
 *  and then returned as meters.
 */
void
DepthOfField::dofInMeters(
    double subjectDistance,
    double &nearDistance,
    double &farDistance
)
{
    if( (circleOfConfusion == 0.0) ||
	(aperture == 0.0) ||
	(focalLength == 0) ) {
	nearDistance = 0.0;
	farDistance = 0.0;
	return;
    }

    /*
     *  Convert subject distance (in meters) to millimetres.
     */
    double distance;
    distance = subjectDistance * 1000.0;

    double hyperfocalDistance =
	    ((double)(focalLength * focalLength) /
	     (aperture * circleOfConfusion)) +
	    (double)focalLength;

    /*
     *  Near limit of acceptable sharpness.
     */
    nearDistance =
	(distance * (hyperfocalDistance - (double)focalLength)) /
	(hyperfocalDistance + distance - 2.0 * (double)focalLength);
    nearDistance = nearDistance / 1000.0;

    /*
     *  Far limit of acceptable sharpness.
     */
    if( (hyperfocalDistance - distance) <= 0.0 ) {
	farDistance = -1.0;	// infinity
    }
    else {
	farDistance =
	    (distance * (hyperfocalDistance - (double)focalLength)) /
	    (hyperfocalDistance - distance);
	farDistance = farDistance / 1000.0;
    }
}

/*
 *  Determine the depth of field in lens positions.
 */
void
DepthOfField::dofInLensPositions(
    int    lensPosition,
    int    &nearPosition,
    int    &farPosition
)
{
    if( (circleOfConfusion == 0.0) ||
	(aperture == 0.0) ||
	(focalLength == 0) ) {
	nearPosition = 0;
	farPosition = 0;
	return;
    }

    double nearDistance;
    double farDistance;

    dofInMeters(
	lens[lensPosition],
	nearDistance,
	farDistance
    );

    int k;
    const int p = lensPosition;

    /*
     *  Determine the lens position on the left (near focus).
     */
    for( k = p; (k > 0) && (lens[k-1] > nearDistance); k-- ) {
	;
    }
    nearPosition = k;

    /*
     *  Determine the lens position on the right (far focus).
     */
    if( farDistance == -1.0 ) {	// infinity
	farPosition = nPos;
    }
    else {
	for( k = p; (k+1 < nPos) && (lens[k+1] < farDistance); k++ ) {
	    ;
	}
	farPosition = k;
    }
}

/*
 *  Determine the hyperfocal distance and resulting depth of
 *  field in lens positions.
 */
void
DepthOfField::hyperfocalInLensPositions(
    int    &lensPosition,
    int    &nearPosition,
    int    &farPosition
)
{
    if( (circleOfConfusion == 0.0) ||
	(aperture == 0.0) ||
	(focalLength == 0) ) {
	lensPosition = 0;
	nearPosition = 0;
	farPosition = 0;
	return;
    }

    /*
     *  Calculation done in millimeters.
     */
    double hyperfocalDistance =
	    ((double)(focalLength * focalLength) /
	     (aperture * circleOfConfusion)) +
	    (double)focalLength;

    int k;

    /*
     *  Convert from millimeters to meters.
     *  Find the corresponding lens position.
     */
    hyperfocalDistance = hyperfocalDistance / 1000.0;
    for( k = 0; (k+1 < nPos) && (lens[k+1] < hyperfocalDistance); k++ ) {
	;
    }
    lensPosition = k;

/*
    printf( "hyperfocalDistance = %12.4f\n",hyperfocalDistance );
    printf( "[%12.4f, infinity]\n", hyperfocalDistance/2.0 );
    for( int j = 3; j <= 9; j += 2 ) {
        printf( "[%12.4f, %12.4f]\n",
	    hyperfocalDistance/(double)(j+1),
	    hyperfocalDistance/(double)(j-1)
        );
    }
*/

    /*
     *  Near limit of acceptable sharpness.
     *  Find the corresponding lens position.
     */
    const int p = lensPosition;
    double nearDistance = hyperfocalDistance / 2.0;
    for( k = p; (k > 0) && (lens[k-1] > nearDistance); k-- ) {
	;
    }
    nearPosition = k;

    /*
     *  Far limit of acceptable sharpness: infinity.
     *  Set the corresponding lens position.
     */
    if( focalLength == 50 ) {
	farPosition = MAX_LensPositions_50mm;
    }
    else { /* focalLength == 200 */
	farPosition = MAX_LensPositions_200mm;
    }
}

