/*
 *  Focus measure: map an image f of size h x w to a value
 *  that represents the degree of focus of the image.
 *
 *  Pixel (i,j) in the image is given by f[i*w + j],
 *  where
 *      i = 0, ..., h-1 (rows)
 *      j = 0, ..., w-1 (columns)
 */

#include <stdio.h>
#include <stdlib.h>
#include "focusMeasure.h"

FocusMeasure::FocusMeasure()
{
}

FocusMeasure::~FocusMeasure()
{
}

/*
 *  Use a first derivative operator to determine a focus measure.
 *
 *  Evaluate the cell with rows y, ..., y + gridHeight-1
 *                  and columns x, ..., x + gridWidth-1.
 */
double
FocusMeasure::squaredGradient( uchar *f, int w, int h, int x, int y, int gridWidth, int gridHeight )
{
    double sum = 0;

    /*
     *  Ignore the boundaries of the image since applying the
     *  operator at, for example, pixel (0,0) is undefined.
     *  I.e., i = 0, ..., h-1 (rows)
     *        j = 0, ..., w-2 (columns)
     */
    for( int i = y; i < y + gridHeight;  i++ )
    for( int j = x; j < x + gridWidth-1; j++ ) {
	int k = i*w + j;
	sum += (f[k+1] - f[k]) * (f[k+1] - f[k]);
	if( sum < 0 ) {
	    fprintf( stderr, "error: overflow in squaredGradient\n" );
	}
    }

    return( sum );
}

/*
 *  Use a first derivative operator to determine a focus measure.
 *
 *  Evaluate the cell with rows y, ..., y + gridHeight-1
 *                  and columns x, ..., x + gridWidth-1.
 */
double
FocusMeasure::brenner( uchar *f, int w, int h, int x, int y, int gridWidth, int gridHeight )
{
    double sum = 0;

    /*
     *  Ignore the boundaries of the image since applying the
     *  operator at, for example, pixel (0,0) is undefined.
     *  I.e., i = 0, ..., h-1 (rows)
     *        j = 0, ..., w-3 (columns)
     */
    for( int i = y; i < y + gridHeight;  i++ )
    for( int j = x; j < x + gridWidth-2; j++ ) {
	int k = i*w + j;
	sum += (f[k+2] - f[k]) * (f[k+2] - f[k]);
	if( sum < 0 ) {
	    fprintf( stderr, "error: overflow in brenner\n" );
	}
    }

    return( sum );
}

/*
 *  Use a second derivative operator to determine a focus measure.
 *  According to Vaquero et al. (2011) this is the operator used in
 *  the Nokia N900 smartphone.
 *
 *  Evaluate the cell with rows y, ..., y + gridHeight-1
 *                  and columns x, ..., x + gridWidth-1.
 */
double
FocusMeasure::nokiaMeasure( uchar *f, int w, int h, int x, int y, int gridWidth, int gridHeight )
{
    double sum = 0;

    /*
     *  Ignore the boundaries of the image since applying the
     *  operator at, for example, pixel (0,0) is undefined.
     *  I.e., i = 0, ..., h-1 (rows)
     *        j = 1, ..., w-2 (columns)
     */
    for( int i = y;     i < y + gridHeight;  i++ )
    for( int j = x + 1; j < x + gridWidth-1; j++ ) {
	int k = i*w + j;
	sum += abs( -1*f[k-1] +  2*f[k] + -1*f[k+1] );
	if( sum < 0 ) {
	    fprintf( stdout, "error: overflow in nokiaMeasure\n" );
	}
    }

    return( sum );
}


/*
 *  Laplacian of Gaussian.
 *  Generated by gaussian.c
 */
double
FocusMeasure::LoG( uchar *f, int w, int h, int x, int y, int gridWidth, int gridHeight )
{
    double sum = 0;

    // remember to adjust range of for loops
    for( int i = y+4; i < y + gridHeight-4;  i++ )
    for( int j = x+4; j < x + gridWidth-4; j++ ) {

        int k = i*w + j;

// sigma =  1.2
double v =
  -0.00604456*f[k-4*w-4] +
  -0.05218375*f[k-4*w-3] +
  -0.22921201*f[k-4*w-2] +
  -0.53574532*f[k-4*w-1] +
  -0.70445656*f[k-4*w+0] +
  -0.53574532*f[k-4*w+1] +
  -0.22921201*f[k-4*w+2] +
  -0.05218375*f[k-4*w+3] +
  -0.00604456*f[k-4*w+4] +

  -0.05218375*f[k-3*w-4] +
  -0.40539537*f[k-3*w-3] +
  -1.53989454*f[k-3*w-2] +
  -3.07029812*f[k-3*w-1] +
  -3.73463936*f[k-3*w+0] +
  -3.07029812*f[k-3*w+1] +
  -1.53989454*f[k-3*w+2] +
  -0.40539537*f[k-3*w+3] +
  -0.05218375*f[k-3*w+4] +

  -0.22921201*f[k-2*w-4] +
  -1.53989454*f[k-2*w-3] +
  -4.42144171*f[k-2*w-2] +
  -5.18823798*f[k-2*w-1] +
  -3.87881214*f[k-2*w+0] +
  -5.18823798*f[k-2*w+1] +
  -4.42144171*f[k-2*w+2] +
  -1.53989454*f[k-2*w+3] +
  -0.22921201*f[k-2*w+4] +

  -0.53574532*f[k-1*w-4] +
  -3.07029812*f[k-1*w-3] +
  -5.18823798*f[k-1*w-2] +
   6.10318853*f[k-1*w-1] +
  18.45137170*f[k-1*w+0] +
   6.10318853*f[k-1*w+1] +
  -5.18823798*f[k-1*w+2] +
  -3.07029812*f[k-1*w+3] +
  -0.53574532*f[k-1*w+4] +

  -0.70445656*f[k+0*w-4] +
  -3.73463936*f[k+0*w-3] +
  -3.87881214*f[k+0*w-2] +
  18.45137170*f[k+0*w-1] +
  40.00000000*f[k+0*w+0] +
  18.45137170*f[k+0*w+1] +
  -3.87881214*f[k+0*w+2] +
  -3.73463936*f[k+0*w+3] +
  -0.70445656*f[k+0*w+4] +

  -0.53574532*f[k+1*w-4] +
  -3.07029812*f[k+1*w-3] +
  -5.18823798*f[k+1*w-2] +
   6.10318853*f[k+1*w-1] +
  18.45137170*f[k+1*w+0] +
   6.10318853*f[k+1*w+1] +
  -5.18823798*f[k+1*w+2] +
  -3.07029812*f[k+1*w+3] +
  -0.53574532*f[k+1*w+4] +

  -0.22921201*f[k+2*w-4] +
  -1.53989454*f[k+2*w-3] +
  -4.42144171*f[k+2*w-2] +
  -5.18823798*f[k+2*w-1] +
  -3.87881214*f[k+2*w+0] +
  -5.18823798*f[k+2*w+1] +
  -4.42144171*f[k+2*w+2] +
  -1.53989454*f[k+2*w+3] +
  -0.22921201*f[k+2*w+4] +

  -0.05218375*f[k+3*w-4] +
  -0.40539537*f[k+3*w-3] +
  -1.53989454*f[k+3*w-2] +
  -3.07029812*f[k+3*w-1] +
  -3.73463936*f[k+3*w+0] +
  -3.07029812*f[k+3*w+1] +
  -1.53989454*f[k+3*w+2] +
  -0.40539537*f[k+3*w+3] +
  -0.05218375*f[k+3*w+4] +

  -0.00604456*f[k+4*w-4] +
  -0.05218375*f[k+4*w-3] +
  -0.22921201*f[k+4*w-2] +
  -0.53574532*f[k+4*w-1] +
  -0.70445656*f[k+4*w+0] +
  -0.53574532*f[k+4*w+1] +
  -0.22921201*f[k+4*w+2] +
  -0.05218375*f[k+4*w+3] +
  -0.00604456*f[k+4*w+4];
/*
*/

	//sum += fabs( v );
	sum += v*v;
    }

    return( sum );
}

