/*
 *  For each global max, print the feature values and the correct
 *  label for that peak.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <algorithm>
#include <vector>
#include <string>
#include "label.h"

using namespace std;

/*
 *  In-focus data for low resolutions images.
 *
 *  typedef struct {
 *    int row;    // row position in 4x6 grid
 *    int col;    // col position in 4x6 grid
 *    int range;  // boolean indicating range of values (1), or a disjuction (0)
 *    int l;  // lower value (of range or disjunction)
 *    int u;  // upper value (of range or disjunction)
 *  } DataType;
 */

#include "../Include/Low/data_backyard.h"
#include "../Include/Low/data_bars.h"
#include "../Include/Low/data_books.h"
#include "../Include/Low/data_building.h"
#include "../Include/Low/data_cans.h"
#include "../Include/Low/data_coins.h"
#include "../Include/Low/data_flowers.h"
#include "../Include/Low/data_trail.h"


Label::Label( int nRows, int nCols, int nImages, int option )
{
    this->nRows = nRows;
    this->nCols = nCols;
    this->nImages = nImages;
    this->option = option;

    /*
     *  Mapping to 4x6 for lookup.
     */
    this->rowFactor = nRows/4;
    this->colFactor = nCols/6;

    depthMap = new int*[nRows];
    for( int i = 0; i < nRows; i++ ) {
	depthMap[i] = new int[nCols];
	for( int j = 0; j < nCols; j++ ) {
	    depthMap[i][j] = -1;
	}
    }
}

Label::~Label()
{
    for( int i = 0; i < nRows; i++ ) {
	delete[] depthMap[i];
    }
    delete[] depthMap;
}

bool
Label::lookup( char *benchmark, int row, int col, int peak, DataType *data )
{
    bool hit = false;
    for( int j = 0; (data[j].row != -1) && !hit; j++ ) {
	hit = (data[j].row == row) &&
	      (data[j].col == col) &&
	      (data[j].l <= peak) && (peak <= data[j].u);
    }

    return( hit );
}

/*
 *  Determine the correct label for a peak.
 */
bool
Label::determineLabel( char *benchmark, int peak, int row, int col )
{
    /*
     *  Map from 32x48 to 4x6 for lookup.
     */
    row = row/rowFactor;
    col = col/colFactor;

    if( strcmp( benchmark, "backyard" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_backyard ) );
    }
    else if( strcmp( benchmark, "bars" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_bars ) );
    }
    else if( strcmp( benchmark, "books" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_books ) );
    }
    else if( strcmp( benchmark, "building" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_building ) );
    }
    else if( strcmp( benchmark, "cans" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_cans ) );
    }
    else if( strcmp( benchmark, "coins" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_coins ) );
    }
    else if( strcmp( benchmark, "flowers" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_flowers ) );
    }
    else if( strcmp( benchmark, "trail" ) == 0 ) {
	return( lookup( benchmark, row, col, peak, data_trail ) );
    }

    return( false );
}

double
Label::getMeasure( int i, int j, int k, double ***measures )
{
    if( (0 <= i) && (i < nRows) &&
        (0 <= j) && (j < nCols) ) {
	return( measures[i][j][k] );
    }
    else {
	return( 0.0 );
    }
}

double
Label::getCellMeasure( int i, int j, int k, double ***measures )
{
    double fm =
	getMeasure( i-1, j-1, k, measures ) +
	getMeasure( i-1, j  , k, measures ) +
	getMeasure( i-1, j+1, k, measures ) +
	getMeasure( i  , j-1, k, measures ) +
	getMeasure( i  , j  , k, measures ) +
	getMeasure( i  , j+1, k, measures ) +
	getMeasure( i+1, j-1, k, measures ) +
	getMeasure( i+1, j  , k, measures ) +
	getMeasure( i+1, j+1, k, measures );

    return( fm );
}

/*
 *  Given the entire depth map and all of the focus measures,
 *  determine feature values and correct label for whether
 *  the depth estimate at depthMap[row][col] is reliable or not.
 */
void
Label::printFeatures( char *benchmark,
		int row, int col, int **depthMap, double ***measures )
{
    int peak = depthMap[row][col];

    double fm[256];

    for( int k = 0; k < nImages; k++ ) {
	if( option == 0 ) fm[k] = measures[row][col][k];
	else		  fm[k] = getCellMeasure( row, col, k, measures );
    }

    /*
     *  Determine the global min and max.
     */
    double min = 1.0;
    double max = 0.0;
    for( int i = 0; i < nImages; i++ ) {
        if( min > fm[i] ) {
            min = fm[i];
        }
        if( max < fm[i] ) {
            max = fm[i];
        }
    }

    /*
     *  Determine the number of peaks.
     */
    int nPeaks = 0;
    for( int i = 1; i < nImages; i++ ) {

	int k;
	for( k = i; k > 0; k-- ) {
	    if( fm[k-1] > fm[k] ) break;
	}
	int height_l = (i - k);

	for( k = i; k < nImages-1; k++ ) {
	    if( fm[k+1] > fm[k] ) break;
	}
	int height_r = (k - i);
	
	int l = 5;
	if( l > i ) l = i;
	int r = 5;
	if( r > (nImages - i) ) r = nImages - i;
	if( (height_l >= l) && (height_r >= r) ) {
	    nPeaks++;
	}
    }

    /*
     *  Determine the mean.
     */
    double mean = 0.0;
    for( int i = 0; i < nImages; i++ ) {
        mean += fm[i];
    }
    mean = mean / nImages;

    /*
     *  Determine the median.
     */
    std::vector <double> window;
    for( int i = 0; i < nImages; i++ ) {
	window.push_back( fm[i] );
    }
    std::nth_element(
	window.begin(),
	window.begin() + window.size()/2,
	window.end() );
    double median = window[window.size()/2];

    /*
     *  Determine the standard deviation, skewness, and kurtosis.
     */
    double standard_deviation = 0.0;
    double skewness = 0.0;
    double kurtosis = 0.0;
    for( int i = 0; i < nImages; i++ ) {
        standard_deviation += pow(fm[i] - mean, 2.0);
        skewness += pow(fm[i] - mean, 3.0);
        kurtosis += pow(fm[i] - mean, 4.0);
    }
    skewness = skewness/nImages;
    kurtosis = kurtosis/nImages;
    skewness = skewness / pow(standard_deviation/nImages, 3.0/2.0);
    kurtosis = kurtosis / pow(standard_deviation/nImages, 2.0);
    standard_deviation = sqrt(standard_deviation / nImages);

    /*
     *  Determine the height on the left.
     */
    int k;
    for( k = 0; (peak-k > 0) && (fm[peak-k] > fm[peak-k-1]); k++ ) {
	;
    }
    int index_l = k;
    double height_l = (fm[peak] - fm[peak-k]);

    /*
     *  Determine the height on the right.
     */
    for( k = 0; (peak+k+1 < nImages) && (fm[peak+k] > fm[peak+k+1]); k++ ) {
	;
    }
    int index_r = k;
    double height_r = (fm[peak] - fm[peak+k]);

    double amplitude;
    if( height_l > height_r ) amplitude = height_l;
    else                      amplitude = height_r;
    int width;
    width = index_r + index_l;
    if( width == 0 ) width = 1;
    double amplitude_over_width = amplitude / (double)width;

    /*
     *  Determine left_ratio_to_lowest_continuous
     *  and right_ratio_to_lowest_continuous.
     */
    double left_ratio_to_lowest_continuous =
	fm[peak-index_l] > 0 ? fm[peak] /fm[peak-index_l] : 0;
    double right_ratio_to_lowest_continuous =
	fm[peak+index_r] > 0 ? fm[peak] /fm[peak+index_r] : 0;

    /*
     *  Determine left_num_in_trend and right_num_in_trend.
     */
    int left_how_many_out_of_10 = 0;
    int prev = peak;
    for (k = 0; (peak-k>0); k++) {
        if (fm[peak-k] < fm[prev]) {
            prev = peak-k;
            left_how_many_out_of_10++;
        }
    }
    int right_how_many_out_of_10 = 0;
    prev = peak;
    for (k = 0; (peak+k<nImages); k++) {
        if (fm[peak+k] < fm[prev]) {
            prev = peak+k;
            right_how_many_out_of_10++;
        }
    }

    /*
     *  Determine left_ratio_to_lowest_out_of_10 and
     *  right_ratio_to_lowest_out_of_10.
     */
    int left_lowest_in_10 = peak;
    for( k = 0; (peak-k > 0) && k < 10; k++ ) {
        if (fm[left_lowest_in_10] > fm[peak-k]) {
            left_lowest_in_10 = peak-k;
        }
    }
    double left_ratio_to_lowest_out_of_10 =
	fm[left_lowest_in_10] > 0 ? fm[peak] / fm[left_lowest_in_10] : 0;
    int right_lowest_in_10 = peak;
    for( k = 0; (peak+k < nImages) && k < 10; k++ ) {
        if (fm[right_lowest_in_10] > fm[peak+k]) {
            right_lowest_in_10 = peak+k;
        }
    }
    double right_ratio_to_lowest_out_of_10 =
	fm[right_lowest_in_10] > 0 ? fm[peak] / fm[right_lowest_in_10] : 0;

    /*
     *  Determine left_ratio_to 3, right_ratio_to 3,
     */
    double left_ratio_to_3 = 0;
    if (peak-3 > 0) {
        left_ratio_to_3 = fm[peak-3] > 0 ? fm[peak] / fm[peak-3] : 0;
    }
    double right_ratio_to_3 = 0;
    if (peak+3 < nImages) {
        right_ratio_to_3 = fm[peak+3] > 0 ? fm[peak] / fm[peak+3] : 0;
    }

    /*
     *  Features with depth map.
     */

    int neighbourDepths[8];
    double avg_surrounding_depth = 0;
    int num_neighbours = 0;

    if(row>0){
        if(col>0){
            avg_surrounding_depth += depthMap[row-1][col-1];
            neighbourDepths[num_neighbours] = depthMap[row-1][col-1];
            num_neighbours++;
        }
        avg_surrounding_depth += depthMap[row-1][col];
        neighbourDepths[num_neighbours] = depthMap[row-1][col];
        num_neighbours++;
        if(col<nCols-1){
            avg_surrounding_depth += depthMap[row-1][col+1];
            neighbourDepths[num_neighbours] = depthMap[row-1][col+1];
            num_neighbours++;
        }
    }
    if(col>0){
        avg_surrounding_depth += depthMap[row][col-1];
        neighbourDepths[num_neighbours] = depthMap[row][col-1];
        num_neighbours++;
    }
    if(col<nCols-1){
        avg_surrounding_depth += depthMap[row][col+1];
        neighbourDepths[num_neighbours] = depthMap[row][col+1];
        num_neighbours++;
    }
    if(row<nRows-1){
        if(col>0){
            avg_surrounding_depth += depthMap[row+1][col-1];
            neighbourDepths[num_neighbours] = depthMap[row+1][col-1];
            num_neighbours++;
        }
        avg_surrounding_depth += depthMap[row+1][col];
        neighbourDepths[num_neighbours] = depthMap[row+1][col];
        num_neighbours++;
        if(col<nCols-1){
            avg_surrounding_depth += depthMap[row+1][col+1];
            neighbourDepths[num_neighbours] = depthMap[row+1][col+1];
            num_neighbours++;
        }
    }
    avg_surrounding_depth/=num_neighbours;
    //double depth_to_avg_surrounding_depth = depthMap[row][col]/avg_surrounding_depth;
    double depth_diff_avg_surrounding_depth = (depthMap[row][col] - avg_surrounding_depth)/avg_surrounding_depth;

    sort(neighbourDepths,neighbourDepths+num_neighbours);
    double medianNeighbourDepth;
    if(num_neighbours%2==1){
        medianNeighbourDepth = 1.0*neighbourDepths[num_neighbours/2];
    }
    else{
        medianNeighbourDepth = (neighbourDepths[num_neighbours/2] + neighbourDepths[num_neighbours/2-1]) / 2.0;
    }

    double depth_to_median_neighbour_depth = depthMap[row][col]/medianNeighbourDepth;

    double depth_standard_deviation = 0.0;
    double depth_kurtosis = 0.0;
    double depth_skewness = 0.0;
    double depth_geometric_mean = 1.0;
    double depth_harmonic_mean = 0.0;
    //double depth_root_mean_square = 0.0;
    double depth_cubic_mean = 0.0;
    for(int i=0;i<num_neighbours;i++){
        depth_standard_deviation += pow(neighbourDepths[i] - avg_surrounding_depth, 2.0);
        depth_skewness += pow(neighbourDepths[i] - avg_surrounding_depth, 3.0);
        depth_kurtosis += pow(neighbourDepths[i] - avg_surrounding_depth, 4.0);
        depth_geometric_mean *= pow(neighbourDepths[i],1.0/num_neighbours);
        depth_harmonic_mean += 1.0/neighbourDepths[i];
        //depth_root_mean_square+= pow(neighbourDepths[i],2);
        depth_cubic_mean+= pow(neighbourDepths[i],3);
    }
    depth_cubic_mean /=num_neighbours;
    depth_cubic_mean = pow(depth_cubic_mean,1.0/3.0);
    //depth_root_mean_square /= num_neighbours;
    //depth_root_mean_square = pow(depth_root_mean_square,0.5);
    depth_skewness = depth_skewness/num_neighbours;
    depth_kurtosis = depth_kurtosis/num_neighbours;
    if(depth_standard_deviation == 0.0){
        depth_skewness = 0.0;
        depth_kurtosis = 0.0;
    }else{
        depth_skewness = depth_skewness / pow(depth_standard_deviation/num_neighbours, 3.0/2.0);
        depth_kurtosis = depth_kurtosis / pow(depth_standard_deviation/num_neighbours, 2.0);
    }
    
    depth_standard_deviation = sqrt(depth_standard_deviation / num_neighbours);

    depth_harmonic_mean = num_neighbours / depth_harmonic_mean;

    //bool less_than_all_neighbours = depthMap[row][col] < neighbourDepths[0];
    //bool greater_than_all_neighbours = depthMap[row][col] > neighbourDepths[num_neighbours-1];
    double same_in_neighbours = 0.0;
    double depth_threshold = 0.11;
    double moving_depth_threshold = 0.0001;
    double withinNeighbourThreshold = 0.0;

    double relative_rank_ratio = 0.0;
    for(int i=0;i<num_neighbours;i++){
        if(depthMap[row][col]>neighbourDepths[i]){
            relative_rank_ratio = relative_rank_ratio + 1.0;
        }
        if(depthMap[row][col]==neighbourDepths[i]){
            same_in_neighbours+=1.0;
        }
        if((double)(abs(depthMap[row][col]-neighbourDepths[i]))/(double)(depthMap[row][col]) <=depth_threshold){
            withinNeighbourThreshold+=1.0;
        }
    }

    same_in_neighbours/=num_neighbours;
    withinNeighbourThreshold/=num_neighbours;
    relative_rank_ratio/=num_neighbours;

    // int depth_range = neighbourDepths[num_neighbours-1] - neighbourDepths[0];

    // int depth_change = std::max(abs(depthMap[row][col] - neighbourDepths[0]), abs(depthMap[row][col] - neighbourDepths[num_neighbours-1]));

    
    int withinRowThreshold = 0;
    int withinMovingRowThreshold = 0;
    int sameInRowCount = 0;
    double rowDepthDiff = 0.0;
    for(int i=0;i<nCols;i++){
        if(depthMap[row][i]==depthMap[row][col]){
            sameInRowCount++;
        }
        //double depth_diff = (double)(abs(depthMap[row][i]-depthMap[row][col]))/(double)(depthMap[row][col]);
        int depth_diff = abs(depthMap[row][i]-depthMap[row][col]);
        //if( depth_diff <=depth_threshold){
        if( depth_diff <= 1 ){
            withinRowThreshold++;
        }
        if( depth_diff <= abs(i-col)*moving_depth_threshold){
            withinMovingRowThreshold++;
        }
        rowDepthDiff+=pow(depthMap[row][i]-depthMap[row][col],2);
    }

    rowDepthDiff/=nCols;
    rowDepthDiff = sqrt(rowDepthDiff);

    double colDepthDiff = 0.0;
    int withinColThreshold = 0;
    int withinMovingColThreshold = 0;
    int sameInColCount = 0;
    for(int i=0;i<nRows;i++){
        if(depthMap[i][col]==depthMap[row][col]){
            sameInColCount++;
        }
        //double depth_diff = (double)(abs(depthMap[i][col]-depthMap[row][col]))/(double)(depthMap[row][col]);
        int depth_diff = abs(depthMap[i][col]-depthMap[row][col]);
        if( depth_diff <= 1 ){
            withinColThreshold++;
        }
        if( depth_diff<= abs(i-row)*moving_depth_threshold){
            withinMovingColThreshold++;
        }
        colDepthDiff+=pow(depthMap[i][col]-depthMap[row][col],2);
    }

    colDepthDiff/=nRows;
    colDepthDiff = sqrt(colDepthDiff);

    int withinLeftDiagThreshold = 0;
    int sameInLeftDiagonal = 0;
    int x = row;
    int y = col;
    while(true){
        x-=1;
        y-=1;
        if(x<0 || y<0){
            break;
        }
        if(depthMap[x][y]==depthMap[row][col]){
            sameInLeftDiagonal++;
        }
        if((double)(abs(depthMap[x][y]-depthMap[row][col]))/(double)(depthMap[row][col]) <=depth_threshold){
            withinLeftDiagThreshold ++;
        }
    }
    x = row;
    y = col;
    while(true){
        x+=1;
        y+=1;
        if(x>=nRows || y>=nCols){
            break;
        }
        if(depthMap[x][y]==depthMap[row][col]){
            sameInLeftDiagonal++;
        }
        if((double)(abs(depthMap[x][y]-depthMap[row][col]))/(double)(depthMap[row][col]) <=depth_threshold){
            withinLeftDiagThreshold ++;
        }
    }

    int withinRightDiagThreshold = 0;
    int sameInRightDiagonal = 0;
    x = row;
    y = col;
    while(true){
        x-=1;
        y+=1;
        if(x<0 || y>=nCols){
            break;
        }
        if(depthMap[x][y]==depthMap[row][col]){
            sameInRightDiagonal++;
        }
        if((double)(abs(depthMap[x][y]-depthMap[row][col]))/(double)(depthMap[row][col]) <=depth_threshold){
            withinRightDiagThreshold ++;
        }
    }
    x = row;
    y = col;
    while(true){
        x+=1;
        y-=1;
        if(x>=nRows || y<0){
            break;
        }
        if(depthMap[x][y]==depthMap[row][col]){
            sameInRightDiagonal++;
        }
        if((double)(abs(depthMap[x][y]-depthMap[row][col]))/(double)(depthMap[row][col]) <=depth_threshold){
            withinRightDiagThreshold ++;
        }
    }

/*
    int withinDepthThreshold = 0;
    int sameDepth = 0;
    for(int i=0;i<nRows;i++){
        for(int j=0;j<nCols;j++){
            if(depthMap[i][j] == depthMap[row][col]){
                sameDepth++;
            }
            if((double)(abs(depthMap[i][j]-depthMap[row][col]))/(double)(depthMap[row][col]) <=depth_threshold){
                withinDepthThreshold ++;
            }
        }
    }
*/

    printf( "\"%s\",%d,%d,%d,%d,",
	benchmark,
	row,
	col,
	peak,
	nPeaks > 0  // unimodal
        //index_l,
        //height_l,
        //index_r,
        //height_r
    );

    printf( "%8.6f,%d,%8.6f,%8.6f,%8.6f,%d,%8.6f,%8.f,",
         left_ratio_to_lowest_continuous,
         left_how_many_out_of_10,
         left_ratio_to_lowest_out_of_10,
         left_ratio_to_3,
         right_ratio_to_lowest_continuous,
         right_how_many_out_of_10,
         right_ratio_to_lowest_out_of_10,
         right_ratio_to_3
     );

    printf( "%8.6f,%8.6f,%8.6f,%8.6f,%8.6f,",
         standard_deviation,
         mean,
         median,
         skewness,
         kurtosis
    );
    //printf( "%8.6f,", (max - min) / (max + min) );
    printf( "%d,", width );
    printf( "%8.6f,", amplitude_over_width );
    printf( "%8.6f,", amplitude );
    printf( "%8.6f,", max - min );
    printf( "%8.6f,", max - mean );
    printf( "%8.6f,", max - median );
    //printf( "%8.6f,", depth_to_avg_surrounding_depth);
    //printf( "%8.6f,",avg_surrounding_depth);
    printf( "%8.6f,",depth_diff_avg_surrounding_depth);
    printf( "%8.6f,",depth_to_median_neighbour_depth);
    //printf( "%8.6f,",(double)sameDepth/((double)nRows*nCols));
    //printf( "%d,",withinDepthThreshold);
    printf( "%8.6f,",(double)sameInRowCount/(double)nCols);
    // printf( "%8.6f,",rowDepthDiff);
    printf( "%8.6f,",(double)sameInColCount/(double)nRows);
    // printf( "%8.6f,",colDepthDiff);
    printf( "%8.6f,",(double)withinRowThreshold/(double)nCols);
    //printf( "%d,",withinMovingRowThreshold);
    printf( "%8.6f,",(double)withinColThreshold/(double)nRows);
    //printf( "%d,",withinMovingColThreshold);
    printf( "%d,",sameInLeftDiagonal);
    printf( "%d,",sameInRightDiagonal);
    printf( "%d,",withinLeftDiagThreshold);
    printf( "%d,",withinRightDiagThreshold);
    //printf( "%8.6f,",same_in_neighbours);
    printf( "%8.6f,",withinNeighbourThreshold);
    //printf( "%8.6f,",depth_root_mean_square);
    // printf( "%8.6f,",depth_cubic_mean);
    // printf( "%8.6f,",depth_geometric_mean);
    // printf( "%8.6f,",depth_harmonic_mean);
    // printf( "%8.6f,",depth_standard_deviation);
    // printf( "%8.6f,",depth_skewness);
    // printf( "%8.6f,",depth_kurtosis);
    // printf( "%d,",less_than_all_neighbours);
    // printf( "%d,",greater_than_all_neighbours);
    // printf( "%8.6f,",relative_rank_ratio);
    // printf( "%d,",depth_range);
    // printf( "%d,",depth_change);

    /*
     * yes --- there is an object in sharp focus
     * no  --- there is no object in sharp focus
     */
    bool label = determineLabel( benchmark, peak, row, col );
    if( label ) printf( "yes\n" );
    else	printf( "no\n" );
}

void
Label::generateData( char *benchmark, double ***measures )
{
    /*
     *  Fill in depth map with max of focus measure for that cell.
     */
    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {

	double fm[256];

	for( int k = 0; k < nImages; k++ ) {
	    if( option == 0 ) fm[k] = measures[i][j][k];
	    else	      fm[k] = getCellMeasure( i, j, k, measures );
	}

	double max = 0.0;
	for( int k = 0; k < nImages; k++ ) {
	    if( max < fm[k] ) {
	        max = fm[k];
	    }
	}
	for( int k = 0; k < nImages; k++ ) {
	    if( max == fm[k] ) {
		depthMap[i][j] = k;
		break;	// take first max closest to camera
	    }
	}
    }

    for( int i = 0; i < nRows; i++ )
    for( int j = 0; j < nCols; j++ ) {
	printFeatures( benchmark, i, j, depthMap, measures);
    }
}

